from ._decomposition import pCA, fastICA, incrementalPCA, kernelPCA, latentDirichletAllocation, miniBatchDictionaryLearning, miniBatchSparsePCA, nMF, sparsePCA, sparseCoder, truncatedSVD, factorAnalysis
